"""The base components for making apps from DAGs"""

import streamlit as st
from dagapp.utils import check_configs, get_default_configs
from dagapp.page_funcs import SimplePageFunc


def dag_to_page_name(dag):
    return f'{list(dag.leafs)[0].capitalize()} Calculator'


def get_page_callbacks(dags, page_names, page_factory, configs):
    return [
        page_factory(dag, page_name, **config)
        for dag, page_name, config in zip(dags, page_names, configs)
    ]


def get_pages_specs(dags, page_factory, configs):
    """
    Return a mapping of page names to page callback objects.

    The `configs` argument is an iterable of per-dag config dicts. If a config
    provides `page_name` or `page_title` that string will be used as the page
    name (and as the page title passed to the page factory). Otherwise a name
    is autogenerated from the dag via `dag_to_page_name`.
    """
    page_names = []
    for dag, config in zip(dags, configs):
        # Prefer explicit config keys; fall back to autogenerated title
        page_name = None
        if isinstance(config, dict):
            page_name = config.get('page_name') or config.get('page_title')
        if not page_name:
            page_name = dag_to_page_name(dag)
        page_names.append(page_name)

    page_callbacks = get_page_callbacks(dags, page_names, page_factory, configs)
    return dict(zip(page_names, page_callbacks))


def dag_app(dags, page_factory=SimplePageFunc, configs=None):
    if configs is None:
        configs = get_default_configs(dags)

    check_configs(dags, configs)

    # st.set_page_config(layout="wide")

    pages = get_pages_specs(dags, page_factory, configs)

    st.sidebar.title('Navigation')
    page = st.sidebar.radio('Select your page', tuple(pages.keys()))

    pages[page]()



"""A simple example of making an app from DAGs"""

from dagapp.base import dag_app
from meshed.dag import DAG
from functools import partial


def b(a):
    return 2 ** a


def d(c):
    return 10 - (5 ** c)


def result(b, d):
    return b * d


dags = [
    DAG((b, d, result)),
]

if __name__ == '__main__':
    app = partial(dag_app, dags=dags)
    app()



"""A dagapp to make binary classifier metrics"""

from collections import Counter
import numpy as np
from functools import partial
from typing import Mapping, Iterable

from meshed.dag import DAG
from dagapp.base import dag_app


def _aligned_items(a, b):
    """Yield (k, a_value, b_value) triples for all k that are both a key of a and of b"""
    # reason for casting to dict is to make sure things like pd.Series use the right
    # keys.
    # could also use k in a.keys() etc. to solve this.
    a = dict(a)
    b = dict(b)
    for k in a:
        if k in b:
            yield k, a[k], b[k]


def _dot_product(a, b):
    """
    >>> _dot_product({'a': 1, 'b': 2, 'c': 3}, {'b': 4, 'c': -1, 'd': 'whatever'})
    5
    """
    return sum(ak * bk for _, ak, bk in _aligned_items(a, b))


def classifier_score(
    confusion_count: Mapping[str, int], confusion_value: Mapping[str, int]
):
    """Compute a score for a classifier that produced the `confusion_count`, based on
    the given `confusion_value`.
    Meant to be curried by fixing the confusion_value dict.

    The function is purposely general -- it is not specific to binary classifier
    outcomes, or even any classifier outcomes.
    It simply computes a normalized dot product, depending on the inputs keys to align
    values to multiply and
    considering a missing key as an expression of a null value.
    """
    return _dot_product(confusion_count, confusion_value) / sum(
        confusion_count.values()
    )


def confusion_count(prediction: Iterable[int], truth: Iterable[int], positive: int = 1):
    """Get a dict containing the counts of all combinations of prediction and
    corresponding truth values.

    >>> confusion_count(
    ... [0, 0, 1, 0, 1, 1, 1],
    ... [0, 0, 0, 1, 1, 1, 1]
    ... )
    {'tn': 2, 'fp': 1, 'fn': 1, 'tp': 3}
    """
    confusion = dict(Counter(zip(prediction, truth)))
    keys = list(confusion.keys())
    for key in keys:
        if key[0] == key[1]:
            if key[1] == positive:
                confusion['tp'] = confusion[key]
            else:
                confusion['tn'] = confusion[key]
        else:
            if key[1] == positive:
                confusion['fn'] = confusion[key]
            else:
                confusion['fp'] = confusion[key]
        del confusion[key]

    keys = [key for key in ['tp', 'tn', 'fn', 'fp'] if key not in confusion.keys()]
    for key in keys:
        confusion[key] = 0

    return confusion


def prediction(predict_proba, threshold):
    """Get an array of predictions from thresholding the scores of predict_proba array.

    >>> prediction([0.3, 0.4, 0.5, 0.6, 0.7, 0.8], threshold=0.5)
    array([False, False,  True,  True,  True,  True])
    """
    return np.array(predict_proba) >= threshold


def predict_proba(model, test_X):
    """Get the prediction_proba scores of a model given some test data"""
    return model.predict_proba(test_X)


dags = [DAG((confusion_count, classifier_score))]
configs = [
    dict(
        arg_types=dict(
            prediction='list', truth='list', positive='num', confusion_value='dict',
        ),
    )
]

if __name__ == '__main__':
    app = partial(dag_app, dags=dags, configs=configs)
    app()



"""Different page functions used to turn dags to app"""

import streamlit as st
from i2 import Sig
from dagapp.utils import (
    get_values,
    get_funcs,
    get_nodes,
    display_factory,
    get_from_configs,
    static_factory,
    vector_factory,
)


class BasePageFunc:
    def __init__(self, dag, page_title: str = '', **config):
        self.dag = dag
        self.page_title = page_title
        self.sig = Sig(dag)
        self.configs = config

    def __call__(self):
        if self.page_title:
            st.markdown(f'''## **{self.page_title}**''')
        st.write(Sig(self.dag))


class SimplePageFunc(BasePageFunc):
    def __call__(self):
        if self.page_title:
            st.markdown(f'''## **{self.page_title}**''')

        c1, c2 = st.columns(2)

        c2.graphviz_chart(self.dag.dot_digraph())

        funcs = get_funcs(self.dag)
        nodes = get_nodes(self.dag)
        values = get_values(self.dag, funcs)
        arg_types, ranges = get_from_configs(self.configs)

        display_factory(self.dag, nodes, funcs, values, arg_types, ranges, c1)


class StaticPageFunc(BasePageFunc):
    def __call__(self):
        if self.page_title:
            st.markdown(f'''## **{self.page_title}**''')

        c1, c2 = st.columns(2)

        c2.graphviz_chart(self.dag.dot_digraph())
        funcs = get_funcs(self.dag)
        nodes = get_nodes(self.dag)
        values = get_values(self.dag, funcs)
        arg_types, ranges = get_from_configs(self.configs)

        static_factory(self.dag, nodes, funcs, values, arg_types, ranges, c1)


class VectorizePageFunc(BasePageFunc):
    def __call__(self):
        if self.page_title:
            st.markdown(f'''## **{self.page_title}**''')

        c1, c2 = st.columns(2)

        c2.graphviz_chart(self.dag.dot_digraph())
        funcs = get_funcs(self.dag)
        nodes = get_nodes(self.dag)
        # values = get_values(self.dag, funcs)
        # arg_types, ranges = get_from_configs(self.configs)

        vector_factory(self.dag, nodes, funcs, c1)



"""An example of making an app from DAGs with configs defined"""

from dagapp.base import dag_app
from meshed.dag import DAG
from functools import partial
from dagapp.page_funcs import StaticPageFunc


def partners(
    max_partners: int, cost_per_click: float = 0.2, price_elasticity: int = 120
):
    return int(max_partners - cost_per_click * price_elasticity)


def clicks(partners: int, clicks_per_partner: float):
    return int(partners * clicks_per_partner)


def revenue(clicks: int, cost_per_click: float = 0.2):
    return clicks * cost_per_click


def user_clicks(a: int = 100, b: int = 3, cost_per_click: float = 0.1):
    return a * (b ** cost_per_click)


def rev(user_clicks: float, revenue_per_click: int = 2):
    return user_clicks * revenue_per_click


def cost(user_clicks: float, cost_per_click: float = 0.1):
    return cost_per_click * user_clicks


def profit(cost: float, rev: float):
    return rev - cost


dags = [
    DAG((user_clicks, rev, cost, profit)),
    DAG((partners, clicks, revenue)),
]

configs = [
    dict(
        arg_types=dict(
            a='num', b='num', cost_per_click='num', revenue_per_click='num',
        ),
    ),
    dict(
        arg_types=dict(
            max_partners='slider',
            cost_per_click='slider',
            price_elasticity='slider',
            partners='slider',
            clicks_per_partner='slider',
        ),
        ranges=dict(
            max_partners=[0, 2000],
            cost_per_click=[0.0, 1.0],
            price_elasticity=[0, 200],
            partners=[0, 1500],
            clicks_per_partner=[0.0, 10.0],
            # clicks=[0, 100000],
            # revenue=[0.0, 1000000.0],
        ),
    ),
]

if __name__ == '__main__':
    app = partial(dag_app, dags=dags, page_factory=StaticPageFunc, configs=configs)
    app()



"""Utils"""

import numpy as np
import pandas as pd
import streamlit as st
from typing import Mapping, Iterable
from meshed.itools import successors
from lined import iterize

DFLT_VALS = {
    int: 0,
    float: 0.0,
    Iterable[int]: '0,0,0,0',
    Mapping[str, int]: dict(tp=0, fn=0, fp=0, tn=0),
    bool: False,
}

DFLT_ANNOT_ARGTYPE_MAP = {
    int: 'num',
    float: 'num',
    str: 'text',
    bool: 'bool',
    list: 'list',
    dict: 'dict',
}

ARG_TYPE_WIDGET_MAP = {
    'num': st.number_input,
    'slider': st.slider,
    'double_slider': st.expander,
    'text': st.text_input,
    'list': st.text_input,
    'dict': st.expander,
}


# ------------------------------------ VECTORIZATION  ------------------------------------


def get_vec_input(node):
    """
    Returns a vectorized input defined by a double slider
    """
    return np.linspace(
        start=int(st.session_state[f'{node}_values'][0]),
        stop=int(st.session_state[f'{node}_values'][1]),
        num=int(st.session_state[f'{node}_num']),
        endpoint=True,
    )


def get_args(dag, node, funcs):
    """
    Returns the arguments for a vectorized FuncNode
    """
    args = list()
    for arg in list(funcs[node].sig.names):
        if arg in dag.roots:
            vec_input = get_vec_input(arg)
            if arg in funcs[node].sig.annotations:
                arg_type = str(funcs[node].sig.annotations[arg])
                if arg_type == 'int':
                    args.append(list(vec_input.astype(int)))
                else:
                    args.append(list(vec_input))
            else:
                args.append(list(vec_input))
        else:
            args.append(st.session_state[arg])
    return args


def display_vec_node(node, funcs, args):
    """
    Displays a non-root-node for vectorized input
    """
    func = iterize(funcs[node].func)
    val = list(func(*args))
    st.session_state[node] = val
    st.write(f'{node}: ')
    st.write(pd.DataFrame(st.session_state[node]))


def mk_double_slider(node, st_kwargs, col):
    """
    Create a double slider for a given node
    """
    with col:
        with st.expander(node):
            st.slider(
                'vectorization range',
                min_value=0,
                max_value=100,
                value=(10, 90),
                key=f'{node}_values',
                **st_kwargs,
            )
            st.number_input(
                'num values', min_value=1, value=5, key=f'{node}_num', **st_kwargs
            )
            st.write(pd.DataFrame(get_vec_input(node)).transpose())


def vector_factory(dag, nodes, funcs, col):
    """
    Displays the root nodes of a vectorized DAG as double sliders
    """
    with col:
        for node in dag.sig.names:
            st_kwargs = dict(on_change=update_vec_nodes, args=(dag, nodes, funcs, col),)
            mk_double_slider(node, st_kwargs, col)


def update_vec_nodes(dag, nodes, funcs, col):
    """
    Update non root-nodes for vectorized DAG factory
    """
    with col:
        for node in [node for node in nodes if node not in dag.roots]:
            args = get_args(dag, node, funcs)
            if len(set(map(len, [arg for arg in args]))) == 1:
                display_vec_node(node, funcs, args)
            else:
                break


# ------------------------------------ STATIC NODES ------------------------------------


def get_kwargs(node, funcs):
    """
    Get keyword arguments for a FuncNode
    """
    kwargs = dict()
    for arg in list(funcs[node].sig.names):
        if arg in funcs[node].sig.annotations:
            arg_type = str(funcs[node].sig.annotations[arg])
        else:
            arg_type = str(float)
        if 'typing.Iterable' in arg_type:
            kwargs[arg] = [int(num) for num in st.session_state[arg].split(',')]
        elif 'typing.Mapping' in arg_type:
            kwargs[arg] = dict(
                tp=st.session_state[f'{arg}_tp'],
                fn=st.session_state[f'{arg}_fn'],
                fp=st.session_state[f'{arg}_fp'],
                tn=st.session_state[f'{arg}_tn'],
            )
        else:
            kwargs[arg] = st.session_state[arg]
    return kwargs


def update_static_nodes(dag, nodes, funcs, col):
    """
    Updates the non-root nodes for a static DAG factory
    """
    with col:
        for node in [node for node in nodes if node not in dag.roots]:
            kwargs = get_kwargs(node, funcs)
            val = funcs[node].func(**kwargs)
            if isinstance(val, dict):
                for key in val.keys():
                    st.session_state[f'{node}_{key}'] = val[key]
                with st.expander(node):
                    for key in val.keys():
                        st.write(f"{key}: {st.session_state[f'{node}_{key}']}")
            else:
                st.session_state[node] = val
                st.write(f'{node}: {st.session_state[node]}')


def static_factory(dag, nodes, funcs, values, arg_types, ranges, col):
    """
    Displays the root nodes of a dag
    """
    with col:
        for node in dag.sig.names:
            st_kwargs = dict(
                value=values[node],
                on_change=update_static_nodes,
                args=(dag, nodes, funcs, col),
                key=node,
            )
            display_node(node, arg_types, ranges, values, st_kwargs)


# ------------------------------------ INTERMEDIATE NODES ------------------------------------


def display_node(node, arg_types, ranges, values, st_kwargs):
    """
    Displays the given node based on the argument type
    """
    if node in arg_types:
        if arg_types[node] == 'dict':
            with ARG_TYPE_WIDGET_MAP['dict'](node):
                for condition in values[node].keys():
                    st_kwargs['value'] = values[node][condition]
                    st_kwargs['key'] = f'{node}_{condition}'
                    st.number_input(condition, **st_kwargs)
        elif arg_types[node] == 'slider':
            st_kwargs['min_value'] = ranges[node][0]
            st_kwargs['max_value'] = ranges[node][1]
            st.slider(node, **st_kwargs)
        elif arg_types[node] == 'bool':
            # streamlit.radio does not accept a 'value' kwarg; use checkbox
            # for boolean inputs which supports 'value' and the same
            # on_change/key/args parameters.
            st.checkbox(
                node,
                value=values[node],
                key=st_kwargs.get('key'),
                on_change=st_kwargs.get('on_change'),
                args=st_kwargs.get('args'),
            )
        else:
            widget = ARG_TYPE_WIDGET_MAP[arg_types[node]]
            widget(node, **st_kwargs)
    else:
        st.number_input(node, **st_kwargs)


def display_factory(dag, nodes, funcs, values, arg_types, ranges, col):
    """
    Display the nodes of a dag with number of slider inputs
    """
    with col:
        for node in nodes:
            st_kwargs = dict(
                value=values[node],
                on_change=update_nodes,
                args=(dag, node, funcs),
                key=node,
            )
            display_node(node, arg_types, ranges, values, st_kwargs)
        st.button(
            'Reload DAG from root nodes', on_click=reload_nodes, args=(dag, funcs)
        )


def reload_nodes(dag, funcs):
    """
    Updates all nodes based on the values of the root nodes
    """
    for node in dag.var_nodes:
        if node not in dag.roots:
            args = [st.session_state[arg] for arg in list(funcs[node].sig.names)]
            st.session_state[node] = funcs[node].func(*args)


def update_nodes(dag, node_ch, funcs):
    """
    Updates successors of a changed node
    """
    sub_nodes = [
        node for node in list(successors(dag.graph, node_ch)) if isinstance(node, str)
    ]
    for node in sub_nodes:
        args = [st.session_state[arg] for arg in list(funcs[node].sig.names)]
        st.session_state[node] = funcs[node].func(*args)


# ------------------------------------ STANDARD UTILS ------------------------------------


def get_root_values(dag):
    """
    Returns the default values for all the root nodes found in dag
    """
    root_defaults = dict()
    for name in dag.sig.names:
        if name in dag.sig.defaults:
            dflt = dag.sig.defaults[name]
            if dflt is not None:
                root_defaults[name] = dflt
        elif name in dag.sig.annotations:
            root_defaults[name] = DFLT_VALS[dag.sig.annotations[name]]
        else:
            root_defaults[name] = 0.0
    return root_defaults


def get_func_values(defaults, funcs):
    """
    Returns the default values for all the FuncNodes in funcs using the root defaults in defaults
    """
    func_defaults = defaults
    for name in funcs:
        func_node = funcs[name]
        if set(func_node.sig.names).issubset(set(list(func_defaults.keys()))):
            kwargs = dict()
            for arg_name in func_node.sig.names:
                kwargs[arg_name] = func_defaults[arg_name]
            func_defaults[name] = func_node.func(**kwargs)
    return func_defaults


def get_values(dag, funcs):
    """
    Returns default values for all the nodes found in dag
    """
    root_defaults = get_root_values(dag)
    func_defaults = get_func_values(root_defaults, funcs)
    return func_defaults


def get_nodes(dag):
    """
    Returns the names of all nodes found in dag
    """
    nodes = dag.sig.names
    for node in dag.nodes:
        if node not in nodes and isinstance(node, str):
            nodes.append(node)
    return nodes


def get_funcs(dag):
    """
    Returns the names of all the FuncNodes found in dag
    """
    funcs = dict()
    for func_node in dag.func_nodes:
        funcs[func_node.name[:-1]] = func_node
    return funcs


def get_from_configs(configs):
    """
    Obtains information from user defined configs
    """
    arg_types = configs['arg_types']
    ranges = None if 'ranges' not in configs else configs['ranges']
    return arg_types, ranges


def get_default_configs(dags):
    """
    Returns default configs based on dags if the user did not provide them
    """
    configs = []
    for dag in dags:
        config = {'arg_types': {}}
        for node in dag.roots:
            if node in dag.sig.annotations:
                config['arg_types'][node] = DFLT_ANNOT_ARGTYPE_MAP[
                    dag.sig.annotations[node]
                ]
            else:
                config['arg_types'][node] = 'num'
        configs.append(config)
    return configs


def check_configs(dags, configs):
    """
    Checks the user defined configs to prevent errors
    """
    if len(configs) != len(dags):
        st_error('You need to define configs for all of your DAGs!')
    for dag, config in zip(dags, configs):
        # config should be a dict containing an 'arg_types' mapping
        if not isinstance(config, dict):
            st_error(
                'Each config must be a dict containing "arg_types" and optional "ranges".'
            )

        if 'arg_types' not in config:
            st_error('You need to define an argument type for your root nodes!')
        else:
            # ensure arg_types provides an entry per root node
            if len(config['arg_types']) < len(dag.roots):
                st_error(
                    'You need to define an argument type for all of your root nodes!'
                )

            # if any of the arg types is a slider, ensure ranges are provided
            if 'slider' in config['arg_types'].values():
                if 'ranges' not in config:
                    st_error(
                        'You need to define slider ranges if you want to set slider as an argument type!'
                    )


def st_error(message):
    """
    Raises a streamlit error with the given message
    """
    st.error(message)
    st.stop()



"""
Rent or Buy.

Based on this article:
"""

import math


def calculate_rent_vs_buy(
    years_to_evaluate=30,
    # Default values extracted from the original NYT article's JavaScript
    home_price=500000,
    down_payment_percent=0.20,
    mortgage_interest_rate=0.07,
    property_tax_rate=0.01,
    home_insurance_annual=1500,
    maintenance_percent=0.01,
    closing_cost_percent=0.02,
    selling_cost_percent=0.06,
    annual_rent=24000,
    rent_increase_rate=0.03,
    home_value_growth_rate=0.04,
    investment_return_rate=0.06,
    marginal_tax_rate=0.25,
    loan_term_years=30,
    inflation_rate=0.025,
):
    """
    Calculates and compares the total net present value of buying vs. renting a home
    over a specified number of years, based on the logic from the NYT calculator.

    Args:
        years_to_evaluate (int): The number of years to run the simulation.
        ... (other arguments for financial variables with default values)

    Returns:
        A dictionary containing the total accumulated costs for both scenarios and
        the breakeven year.
    """

    # Initialize variables for both scenarios
    npv_buy_total = 0.0
    npv_rent_total = 0.0
    breakeven_year = None

    # Initial setup for buying scenario
    principal = home_price * (1 - down_payment_percent)
    closing_costs = home_price * closing_cost_percent

    # Amortization formula to calculate monthly mortgage payment
    monthly_interest_rate = mortgage_interest_rate / 12
    num_payments = loan_term_years * 12

    if monthly_interest_rate > 0:
        monthly_mortgage_payment = principal * (
            monthly_interest_rate
            / (1 - math.pow(1 + monthly_interest_rate, -num_payments))
        )
    else:
        # Handle case where interest rate is 0
        monthly_mortgage_payment = principal / num_payments

    current_loan_balance = principal
    total_interest_paid_yearly = 0.0

    # Initial investment for renting scenario (down payment + closing costs)
    initial_rent_investment = (home_price * down_payment_percent) + closing_costs
    rent_investment_value = initial_rent_investment

    # Main simulation loop
    for year in range(1, years_to_evaluate + 1):

        # --- BUYING SCENARIO CALCULATIONS ---

        # Calculate yearly costs
        annual_property_tax = home_price * property_tax_rate
        annual_maintenance = home_price * maintenance_percent
        annual_insurance = home_insurance_annual

        # Calculate mortgage interest paid for the year and update loan balance
        annual_principal_paid = 0.0
        total_interest_paid_yearly = 0.0

        if year <= loan_term_years:
            for _ in range(12):
                interest_this_month = current_loan_balance * monthly_interest_rate
                principal_this_month = monthly_mortgage_payment - interest_this_month
                current_loan_balance -= principal_this_month
                total_interest_paid_yearly += interest_this_month
                annual_principal_paid += principal_this_month

        # Calculate tax deductions on mortgage interest and property taxes
        tax_deduction = (
            total_interest_paid_yearly + annual_property_tax
        ) * marginal_tax_rate

        # Total annual cost of buying before selling and home appreciation
        buy_cost_this_year = (
            (monthly_mortgage_payment * 12)
            + annual_property_tax
            + annual_maintenance
            + annual_insurance
            - tax_deduction
        )

        # Update home value for the year
        home_value_after_year = home_price * math.pow(1 + home_value_growth_rate, year)

        # Calculate selling cost and equity at the end of the simulation
        if year == years_to_evaluate:
            selling_cost = home_value_after_year * selling_cost_percent

            # Calculate total equity gained from principal payments and appreciation
            total_equity = home_value_after_year - current_loan_balance

            # Final costs and gains
            buy_cost_this_year += selling_cost
            buy_cost_this_year -= total_equity

        # Discount annual cost back to present value
        npv_buy_total += buy_cost_this_year / math.pow(1 + inflation_rate, year)

        # --- RENTING SCENARIO CALCULATIONS ---

        # Annual cost of rent
        annual_rent_cost = annual_rent * math.pow(1 + rent_increase_rate, year - 1)

        # Calculate additional funds invested from not buying
        rent_savings_invested = (
            (monthly_mortgage_payment * 12)
            + (home_price * property_tax_rate)
            + (home_price * maintenance_percent)
            + home_insurance_annual
        )

        # The key difference from the original NYT logic is that the invested rent savings
        # are calculated separately and compounded. The provided HTML logic does this
        # on an NPV basis. Let's stick to the NPV model for direct comparison.

        # Discount annual rent cost back to present value
        npv_rent_total += annual_rent_cost / math.pow(1 + inflation_rate, year)

        # Account for investment gains
        # The original script's logic for NPV of investments is complex.
        # This simplified version adds the gains from the initial investment.
        rent_investment_value = rent_investment_value * (1 + investment_return_rate)

        # Update breakeven year
        if breakeven_year is None and npv_buy_total < npv_rent_total:
            breakeven_year = year

    # The original script adds the down payment and closing costs to the final NPV
    # This is a common practice to include initial costs in the total NPV calculation
    npv_buy_total += home_price * down_payment_percent + closing_costs
    npv_rent_total -= rent_investment_value

    return {
        'npv_buy_total': int(npv_buy_total),  # rounding to the nearest int
        'npv_rent_total': int(npv_rent_total),  # rounding to the nearest int
        'breakeven_year': breakeven_year,
    }


if __name__ == '__main__':

    from dagapp.base import dag_app

    from meshed.dag import DAG
    from functools import partial

    rendering_kind = 'static_page_func'

    if rendering_kind == 'static_page_func':
        from dagapp.page_funcs import StaticPageFunc

        app = partial(
            dag_app, dags=[DAG((calculate_rent_vs_buy,))], page_factory=StaticPageFunc
        )
    elif rendering_kind == 'json_page_func' or rendering_kind == 'table_page_func':
        # Custom page function that renders dict outputs as JSON or as a one-row table
        import pandas as pd
        import streamlit as st

        from dagapp.page_funcs import BasePageFunc
        from dagapp.utils import (
            get_funcs,
            get_nodes,
            get_values,
            get_from_configs,
            display_node,
            get_kwargs,
        )

        def format_display_value(v):
            """Format numeric values for display: no decimals and space as thousands separator.

            - ints -> "123 456"
            - floats -> rounded to nearest int and formatted as above
            - iterables -> format each element
            - other types returned unchanged
            """
            # handle numpy scalars/arrays
            try:
                import numpy as _np
            except Exception:
                _np = None

            def fmt_num(x):
                try:
                    if _np is not None and isinstance(x, _np.generic):
                        x = x.item()
                    if isinstance(x, bool):
                        return x
                    if isinstance(x, int):
                        return f'{x:,}'.replace(',', ' ')
                    if isinstance(x, float):
                        # round to nearest integer for display
                        return f'{int(round(x)):,}'.replace(',', ' ')
                except Exception:
                    pass
                return x

            # lists/tuples/ndarrays
            if (
                _np is not None
                and hasattr(_np, 'ndarray')
                and isinstance(v, _np.ndarray)
            ):
                return [fmt_num(x) for x in v.tolist()]
            if isinstance(v, (list, tuple)):
                return [fmt_num(x) for x in v]

            if isinstance(v, dict):
                # format dict values recursively
                return {k: format_display_value(val) for k, val in v.items()}

            return fmt_num(v)

        def update_static_nodes_dict(dag, nodes, funcs, col):
            """Update non-root nodes and render dict outputs as JSON/table."""
            with col:
                for node in [node for node in nodes if node not in dag.roots]:
                    kwargs = get_kwargs(node, funcs)
                    val = funcs[node].func(**kwargs)
                    if isinstance(val, dict):
                        # format values for display
                        disp_val = format_display_value(val)
                        with st.expander(node):
                            if rendering_kind == 'json_page_func':
                                st.json(disp_val)
                            else:
                                # table: render dict as a single-row DataFrame
                                st.table(pd.DataFrame([disp_val]))
                    else:
                        st.session_state[node] = val
                        st.write(f'{node}: {st.session_state[node]}')

        def reload_nodes_dict(dag, funcs):
            for node in dag.var_nodes:
                if node not in dag.roots:
                    args = [
                        st.session_state[arg] for arg in list(funcs[node].sig.names)
                    ]
                    st.session_state[node] = funcs[node].func(*args)

        class DictPageFunc(BasePageFunc):
            def __call__(self):
                if self.page_title:
                    st.markdown(f'''## **{self.page_title}**''')

                c1, c2 = st.columns(2)

                c2.graphviz_chart(self.dag.dot_digraph())
                funcs = get_funcs(self.dag)
                nodes = get_nodes(self.dag)
                values = get_values(self.dag, funcs)
                arg_types, ranges = get_from_configs(self.configs)

                with c1:
                    for node in self.dag.sig.names:
                        st_kwargs = dict(
                            value=values[node],
                            on_change=update_static_nodes_dict,
                            args=(self.dag, nodes, funcs, c1),
                            key=node,
                        )
                        display_node(node, arg_types, ranges, values, st_kwargs)

                    st.button(
                        'Reload DAG from root nodes',
                        on_click=reload_nodes_dict,
                        args=(self.dag, funcs),
                    )

                # initial render
                update_static_nodes_dict(self.dag, nodes, funcs, c1)

        app = partial(
            dag_app, dags=[DAG((calculate_rent_vs_buy,))], page_factory=DictPageFunc
        )
    else:
        raise ValueError(f'Unknown rendering kind: {rendering_kind}')

    app()



"""An example of vectorizing inputs with DAGs"""

from dagapp.base import dag_app
from meshed.dag import DAG
from functools import partial
from dagapp.page_funcs import VectorizePageFunc


def b(a: int):
    return 2 ** a


def d(c: int):
    return 10 - (5 ** c)


def result(b: int, d: int):
    return b * d


dags = [DAG([b, d, result])]

if __name__ == '__main__':
    app = partial(dag_app, dags=dags, page_factory=VectorizePageFunc)
    app()



"""Compute consulting fees

All common parameter defaults are centralized in the `dflt` single-source-of-truth
so functions that share parameter names also share the same defaults. This
avoids signature conflicts when building a DAG using meshed (parameter merging).
"""

from typing import Literal
from types import SimpleNamespace

from dagapp.base import dag_app
from meshed.dag import DAG
from functools import partial

# Single source of truth for shared defaults used across function signatures.
# Use attribute access (dflt.<name>) so defaults are clearly named and grouped.
dflt = SimpleNamespace(
    # billing defaults
    retainer=True,
    ad_hoc_day_rate=2800.0,
    retainer_day_rate=2400.0,
    # time defaults
    hours_per_day=8.0,
    committed_days=8.0,
    ramp_up_days=0.0,
    # percentage / rates
    flex_premium_pct=0.10,
    vat_rate=0.081,
)


def effective_day_rate(
    retainer: bool = dflt.retainer,
    ad_hoc_day_rate: float = dflt.ad_hoc_day_rate,
    retainer_day_rate: float = dflt.retainer_day_rate,
) -> float:
    """Return the effective day-rate based on retainer flag."""
    return retainer_day_rate if retainer else ad_hoc_day_rate


def effective_hourly_rate(
    effective_day_rate: float, hours_per_day: float = dflt.hours_per_day,
) -> float:
    """Return the hourly equivalent of the effective day-rate."""
    return effective_day_rate / hours_per_day


def committed_days(tier: Literal['light', 'regular', 'deep'] = 'regular',) -> float:
    """Return default committed days/month for a named tier."""
    mapping = {'light': 4.0, 'regular': 8.0, 'deep': 12.0}
    return mapping[tier]


def base_fee_ex_vat(
    effective_day_rate: float, committed_days: float = dflt.committed_days,
) -> float:
    """Return the base monthly fee (ex-VAT) for committed days."""
    return effective_day_rate * committed_days


def flex_fee_ex_vat(
    effective_day_rate: float,
    ramp_up_days: float = dflt.ramp_up_days,
    flex_premium_pct: float = dflt.flex_premium_pct,
) -> float:
    """
    Return the flex (short-notice) component (ex-VAT).
    Applies only to additional ramp-up days beyond the commitment.
    """
    return effective_day_rate * ramp_up_days * flex_premium_pct


def subtotal_ex_vat(base_fee_ex_vat: float, flex_fee_ex_vat: float = 0.0,) -> float:
    """Return the subtotal (ex-VAT)."""
    return base_fee_ex_vat + flex_fee_ex_vat


def vat_amount(subtotal_ex_vat: float, vat_rate: float = dflt.vat_rate,) -> float:
    """Return the VAT amount."""
    return subtotal_ex_vat * vat_rate


def total_incl_vat(subtotal_ex_vat: float, vat_amount: float,) -> float:
    """Return the total including VAT."""
    return subtotal_ex_vat + vat_amount


# --- Convenience wrappers for common scenarios ---


def light_monthly_total_incl_vat(
    retainer: bool = dflt.retainer,
    ad_hoc_day_rate: float = dflt.ad_hoc_day_rate,
    retainer_day_rate: float = dflt.retainer_day_rate,
    ramp_up_days: float = dflt.ramp_up_days,
    flex_premium_pct: float = dflt.flex_premium_pct,
    vat_rate: float = dflt.vat_rate,
) -> float:
    """Return total incl. VAT for the 'light' (4 d/mo) scenario."""
    d = effective_day_rate(retainer, ad_hoc_day_rate, retainer_day_rate)
    base = base_fee_ex_vat(d, committed_days('light'))
    flex = flex_fee_ex_vat(d, ramp_up_days, flex_premium_pct)
    sub = subtotal_ex_vat(base, flex)
    vat = vat_amount(sub, vat_rate)
    return total_incl_vat(sub, vat)


def regular_monthly_total_incl_vat(
    retainer: bool = dflt.retainer,
    ad_hoc_day_rate: float = dflt.ad_hoc_day_rate,
    retainer_day_rate: float = dflt.retainer_day_rate,
    ramp_up_days: float = dflt.ramp_up_days,
    flex_premium_pct: float = dflt.flex_premium_pct,
    vat_rate: float = dflt.vat_rate,
) -> float:
    """Return total incl. VAT for the 'regular' (8 d/mo) scenario."""
    d = effective_day_rate(retainer, ad_hoc_day_rate, retainer_day_rate)
    base = base_fee_ex_vat(d, committed_days('regular'))
    flex = flex_fee_ex_vat(d, ramp_up_days, flex_premium_pct)
    sub = subtotal_ex_vat(base, flex)
    vat = vat_amount(sub, vat_rate)
    return total_incl_vat(sub, vat)


def deep_monthly_total_incl_vat(
    retainer: bool = dflt.retainer,
    ad_hoc_day_rate: float = dflt.ad_hoc_day_rate,
    retainer_day_rate: float = dflt.retainer_day_rate,
    ramp_up_days: float = dflt.ramp_up_days,
    flex_premium_pct: float = dflt.flex_premium_pct,
    vat_rate: float = dflt.vat_rate,
) -> float:
    """Return total incl. VAT for the 'deep' (12 d/mo) scenario."""
    d = effective_day_rate(retainer, ad_hoc_day_rate, retainer_day_rate)
    base = base_fee_ex_vat(d, committed_days('deep'))
    flex = flex_fee_ex_vat(d, ramp_up_days, flex_premium_pct)
    sub = subtotal_ex_vat(base, flex)
    vat = vat_amount(sub, vat_rate)
    return total_incl_vat(sub, vat)


# --- One-shot calculator when you just want the numbers quickly ---


def monthly_totals(
    committed_days: float = dflt.committed_days,
    ramp_up_days: float = dflt.ramp_up_days,
    retainer: bool = dflt.retainer,
    ad_hoc_day_rate: float = dflt.ad_hoc_day_rate,
    retainer_day_rate: float = dflt.retainer_day_rate,
    flex_premium_pct: float = dflt.flex_premium_pct,
    vat_rate: float = dflt.vat_rate,
) -> dict:
    """
    Return a dict with all intermediate values for transparency:
    effective_day_rate, base_fee_ex_vat, flex_fee_ex_vat, subtotal_ex_vat,
    vat_amount, total_incl_vat.
    """
    d = effective_day_rate(retainer, ad_hoc_day_rate, retainer_day_rate)
    base = base_fee_ex_vat(d, committed_days)
    flex = flex_fee_ex_vat(d, ramp_up_days, flex_premium_pct)
    sub = subtotal_ex_vat(base, flex)
    vat = vat_amount(sub, vat_rate)
    total = total_incl_vat(sub, vat)
    return {
        'effective_day_rate': d,
        'base_fee_ex_vat': base,
        'flex_fee_ex_vat': flex,
        'subtotal_ex_vat': sub,
        'vat_amount': vat,
        'total_incl_vat': total,
    }


def get_dag():
    funcs = {
        # "monthly_totals": monthly_totals,
        # "deep_monthly_total_incl_vat": deep_monthly_total_incl_vat,
        # "regular_monthly_total_incl_vat": regular_monthly_total_incl_vat,
        # "light_monthly_total_incl_vat": light_monthly_total_incl_vat,
        # "committed_days": committed_days,
        # "effective_hourly_rate": effective_hourly_rate,
        'effective_day_rate': effective_day_rate,
        'base_fee_ex_vat': base_fee_ex_vat,
        'flex_fee_ex_vat': flex_fee_ex_vat,
        'subtotal_ex_vat': subtotal_ex_vat,
        'vat_amount': vat_amount,
        'total_incl_vat': total_incl_vat,
    }

    dag = DAG(funcs.values())

    return dag


if __name__ == '__main__':
    dag = get_dag()
    collapsed_dag = DAG([dag], name='simple')
    dags = [collapsed_dag, dag]

    # configs = [
    #     {"page_name": "simple"},  # for collapsed_dag
    #     {"page_name": "details"},  # for dag
    # ]
    configs = None

    app = partial(dag_app, dags=dags, configs=configs)
    app()
